/*  plumb - the pipe shell - create an instance of a subscript
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "plumb.h"
#include "_default.h"
#include "pipe.h"
#include "vp.h"
#include "parse_common.h"
#include "var.h"
#include "virt.h"

typedef struct {
	transaction_t *tx;
	proc_virt_t *pe;
} new_t;

/* Trickery with new->tx: from init() create everything from the transaction
   but do not run processes yet. These processes kept in stopped mode will
   let new_start() rebind all the env fds, and then new_start() can run
   all processes on the tx. */

static virt_ctrl_t new_start(proc_virt_t *p)
{
	int n;
	fd_t *locfd, *remfd;
	new_t *nw = p->data;
	proc_virt_t *pe = nw->pe; /* sub-env process */

	for(n = 0; n < pe->h.num_fds; n++) {
		if (!vp_is_bound(pe, n))
			continue;
		if ((n >= p->h.num_fds) || (!vp_is_bound(p, n)))
			plumb_error("Can not start [new] %s because fd %d is not bound.\n", p->h.id, n);
		if (is_source(&pe->h.fds[n])) {
			/* sub.env source fd */
			locfd = pe->h.fds[n].pipe->to;   /* local sink */
			remfd = p->h.fds[n].pipe->from;  /* remote source */
			pipe_rebind(p->h.fds[n].pipe, remfd, locfd);

		}
		else if (is_sink(&pe->h.fds[n])) {
			/* sub.env source fd */
			locfd = pe->h.fds[n].pipe->from;   /* local source */
			remfd = p->h.fds[n].pipe->to;      /* remote sink */
			pipe_rebind(p->h.fds[n].pipe, locfd, remfd);
		}
	}

	tx_finish_run(nw->tx);
	nw->tx = NULL;
	proc_free(pe);
	nw->pe = NULL;
	return VC_OK;
}

static void new_free(proc_virt_t *p)
{
	new_t *nw = p->data;

	if (nw->tx != NULL)
		tx_discard(nw->tx);

	if (nw->pe != NULL)
		proc_free(nw->pe);
	free(nw);
}


int vp_new_init(proc_virt_t *p, kvlist_t *args)
{
	const char *orig_script;
	char *pe_id, *inp_name;
	kv_t *parm;
	new_t *nw;
	int idlen;

	/* allocate a process descriptor */
	p->input = NULL;
	p->sink_eof = NULL;
	p->source_eof = NULL;
	p->source_block = NULL;
	p->start = new_start;
	p->free = new_free;
	p->data = nw = malloc(sizeof(new_t));

	if (args->len < 1)
		plumb_error("[new] needs at least one argument");

	idlen = strlen(p->h.id);
	pe_id = malloc(idlen+8);
	sprintf(pe_id, "%s.env", p->h.id);
	nw->pe = proc_create(p->tx, pe_id, 1, NULL, NULL, NULL);


	parm = args->head;
	orig_script = var_get(p->tx, parm->key);

	if (orig_script == NULL)
		plumb_error("Can't [new] variable %s: does not exist\n", parm->key);

	inp_name = malloc(strlen(p->h.id) + strlen(parm->key) + 64);
	sprintf(inp_name, "%s=[new %s ...]", p->h.id, parm->key);

	/* set vars saving their old values as needed */
	nw->tx = tx_this_or_new(NULL, inp_name, fd_ref_(p, vstdout));
	nw->tx->err_on_flush = malloc(idlen+64);
	sprintf(nw->tx->err_on_flush, "command \"flush\" is invalid in a %s=[new] context\n", p->h.id);
	for(parm = args->head->next; parm != NULL; parm = parm->next) {
		var_t *v;
		v = var_set(nw->tx, parm->key, parm->value);
		v->flags |= VAR_NOFLUSH;
	}

	plumb_parse_string(nw->tx, p->h.id, inp_name, orig_script, fd_ref_(p, vstdout));
	tx_finish_create(nw->tx);

	free(inp_name);
	return 0;
}
